"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

from __future__ import annotations

from dataclasses import dataclass
import datetime
import zoneinfo

import ics  # type: ignore

from . import Date, Dates
from ..utils import data2hash


@dataclass
class CourtDate(Date):
    """
    Holds court date
    """

    date: str
    when: str
    where: str
    what: str


class CourtDates(Dates):
    """
    Holds court dates
    """

    # Court dates
    data: list[CourtDate]

    # Sort order
    sort_order: list[str] = ["date", "name2sort", "when", "where", "what"]

    def __init__(self, court_dates: list[dict[str, str]]) -> None:
        """
        Creates 'CourtDates' instance

        :param court_dates: list[dict[str, str]] Data representing court dates
        :return: None
        """

        self.data = [CourtDate(**item) for item in court_dates]

    def filter(self, query: list[str] | str) -> CourtDates:
        """
        Filters court dates by search term(s)

        :param query: List[str] | str Search terms
        :return: CourtDates Filtered court dates
        """

        return CourtDates(super().filter(query))

    def data2ics(self, duration: int = 1) -> ics.Calendar:
        """
        Exports court dates as ICS calendar object

        :param duration: int Duration of each assignment (in hours)
        :return: ics.Calendar
        """

        # Create calendar
        calendar = ics.Calendar(creator=self.creator)

        # Define timezone
        timezone = zoneinfo.ZoneInfo(self.timezone)

        # Iterate over assignments
        for item in self.data:
            # Define timezone, date & times
            time = self.to_time(item.date + item.when, "%Y-%m-%d%H:%M")
            begin = time.replace(tzinfo=timezone)
            end = begin + datetime.timedelta(hours=duration)

            # Create event
            event = ics.Event(
                uid=data2hash(item),
                name=f"Sitzungsdienst ({item.what})",
                created=datetime.datetime.now(timezone),
                begin=begin,
                end=end,
                location=item.where,
            )

            # Add assignee(s) as attendee(s)
            self.add_attendees(item, event)

            # Add event to calendar
            calendar.events.add(event)

        return calendar



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

import asyncio
import io
import re

from PyPDF2 import PdfReader
from PyPDF2._utils import StrByteType

from .models.court import CourtDates
from .models.express import ExpressDates
from .regex import get_court_dates, get_express_dates, get_people
from .utils import data2hash

# Define input data type
Data = StrByteType | bytes


class StA:
    """
    Handles weekly PDF assignments as issued by 'web.sta'
    """

    @staticmethod
    async def run(pdf: list[Data] | Data) -> tuple[CourtDates, ExpressDates]:
        """
        Runs the whole thing - party time!

        :param pdf: list[Data] | Data PDF contents OR filepath(s)
        :return: tuple[CourtDates, ExpressDates]
        """

        # If not list ..
        if not isinstance(pdf, list):
            # .. convert it
            pdf = [pdf]

        # Create data arrays
        court_dates: list[dict[str, str]] = []
        express_dates: list[dict[str, str]] = []

        # Initialize locks
        court_lock = asyncio.Lock()
        dedupe_lock = asyncio.Lock()
        express_lock = asyncio.Lock()

        # Create hash array
        hashes = set()

        # Loop over input data
        for item in pdf:
            # If necessary ..
            if isinstance(item, bytes):
                # .. convert input data
                item = io.BytesIO(item)

            # Create data array
            pages: dict[int, list[str]] = {}

            # Browse pages
            for idx, page in enumerate(PdfReader(item).pages):
                # Retrieve PDF pages
                page: list[str] = [
                    text.strip() for text in page.extract_text().splitlines() if text
                ]

                # If first page ..
                if idx == 0:
                    # .. retrieve express service dates
                    for item in await StA.express(page):
                        # Calculate hash over data
                        express_hash = data2hash(item)

                        # If not processed before ..
                        if express_hash not in hashes:
                            # .. add it
                            # (1) Aquire locks
                            # (2) Store data
                            async with dedupe_lock:
                                hashes.add(express_hash)

                            async with express_lock:
                                express_dates.append(item)

                pages[idx] = page

            # Retrieve court dates
            for item in await StA.process(await StA.extract(pages)):
                # Calculate hash over data
                court_hash = data2hash(item)

                # If not processed before ..
                if court_hash not in hashes:
                    # .. add it
                    # (1) Aquire locks
                    # (2) Store data
                    async with dedupe_lock:
                        hashes.add(court_hash)

                    # Aquire lock
                    async with court_lock:
                        # Store data
                        court_dates.append(item)

        # Remove duplicates
        return CourtDates(court_dates), ExpressDates(express_dates)

    @staticmethod
    def runs(pdf: list[Data] | Data) -> tuple[CourtDates, ExpressDates]:
        """
        Like 'run', but synchronous (not really)

        :param pdf: list[Data] | Data PDF contents OR filepath(s)
        :return: tuple[CourtDates, ExpressDates]
        """

        return asyncio.run(StA.run(pdf))

    @staticmethod
    async def extract(pages: dict[int, list[str]]) -> dict[str, dict[int, list[str]]]:
        """
        Extracts raw data from PDF pages

        :param pages: dict[int, list[str]] PDF pages
        :return: dict Raw source data
        """

        # Create data array
        raw: dict[str, list[str]] = {}

        # Reset weekday buffer
        date: str = ""

        # Extract assignment data
        for page in pages.values():
            # Reset mode
            is_live = False

            for index, text in enumerate(page):
                # Determine starting point ..
                if text == "Anfahrt":
                    is_live = True

                    # .. and proceed with next entry
                    continue

                # Determine terminal point ..
                if text == "Seite":
                    is_live = False

                    # .. and proceed with next entry
                    continue

                # Enforce entries between starting & terminal point
                if not is_live or "Ende der Auflistung" in text:
                    continue

                # Determine current date / weekday
                if text in ["Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag"]:
                    date = page[index + 1]

                    if date not in raw:
                        raw[date] = []

                    # Proceed with next entry
                    continue

                # Proceed with next entry if it indicates ..
                # (1) .. current date
                if text == date:
                    continue

                # (2) .. follow-up appointment for main trial
                if text in ["F", "+"]:
                    continue

                raw[date].append(text)

        return raw

    @staticmethod
    async def process(raw: dict[str, list[str]]) -> list[dict[str, str]]:
        """
        Processes raw data

        :param raw: dict Raw data
        :return: list[dict[str, str]] Data representing court dates
        """

        # Create data array
        unprocessed: list[tuple[str, str, list[str]]] = []

        # Iterate over source data
        for date, item in raw.items():
            buffer: list[str] = []
            court: str = ""

            # Iterate over text blocks
            for index, text in enumerate(item):
                if is_court(text):
                    court = text

                else:
                    buffer.append(text)

                if index + 1 == len(item) or is_court(item[index + 1]):
                    unprocessed.append((date, court, buffer))

                    # Reset buffer
                    buffer = []

        # Reset global data array
        data: list[dict[str, str]] = []

        for item in unprocessed:
            # Unpack values
            date, court, buffer = item

            # Format data as string
            string = " ".join(buffer)

            # Loop over court dates
            for court_date in get_court_dates(string):
                # Retrieve raw data on assigned people
                people = get_people(court_date["assigned"])

                # Loop over them
                for person in people:
                    # Store data
                    data.append(
                        {
                            "who": person,
                            "date": format_date(date),
                            "when": court_date["time"],
                            "where": format_place(court, court_date["where"]),
                            "what": court_date["docket"],
                            "others": [item for item in people if item != person],
                        }
                    )

        return data

    @staticmethod
    async def express(page: list[str]) -> list[dict[str, str]]:
        """
        Extracts express service dates

        :param page: list[str] PDF page
        :return: list[dict[str, str]] Data representing express service dates
        """

        # Create data array
        express: list[str] = []

        # Detect 'express mode'
        # (1) Activation
        is_express = False

        for text in page:
            # Skip if no express service
            if text == "Keine Einteilung":
                break

            # Determine express service ..
            if text == "Eildienst":
                is_express = True

                # .. and proceed with next entry
                continue

            # Skip
            if text == "Tag":
                break

            if is_express:
                express.append(text)

        # Combine data to string for easier regEx matching
        string = " ".join(express)

        # Create data buffer
        data: list[dict[str, str]] = []

        # Loop over express service dates
        for express_date in get_express_dates(string):
            # Retrieve raw data on assigned people
            people = get_people(express_date["assigned"])

            # Loop over them
            for person in people:
                # Store data
                data.append(
                    {
                        "start": format_date(express_date["start"]),
                        "end": format_date(express_date["end"]),
                        "who": person,
                        "others": [item for item in people if item != person],
                    }
                )

        return data


def is_court(string: str) -> bool:
    """
    Checks whether string denotes district or regional court

    :param string: str String to be checked
    :return: bool Whether or not string denotes a court
    """

    if re.match(r"(?:AG|LG)\s", string.strip()):
        return True

    return False


def format_date(string: str) -> str:
    """
    Formats (german) date format DIN 1355-1 to ISO 8601

    :param string: str String representing date
    :return: str Formatted date
    """

    return "-".join(reversed(string.split(".")))


def format_place(court: str, extra: str) -> str:
    """
    Formats court & additional information

    :param court: str String representing a court
    :param extra: str String holding additional information
    :return: str Formatted place
    """

    # Format string representing court
    string = court.replace(" ,", "").strip()

    return f"{string} {extra.strip()}" if extra else string



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import copy
from dataclasses import asdict, dataclass
import datetime
from operator import attrgetter
from typing import Any, Sequence

import ics  # type: ignore


@dataclass
class Date:
    """
    Base class for single assignment
    """

    # Assigned person
    who: dict[str, str]

    # Assigned people (other than 'who')
    others: list[dict[str, str]]

    def to_dict(self) -> dict[str, Any]:
        """
        Exports data

        :return: dict[str, Any]
        """

        return asdict(self)

    @property
    def name(self) -> str:
        """
        Builds full name (including title & department)

        :return: str Full name
        """

        return format_person(self.who)

    @property
    def name2sort(self) -> str:
        """
        Provides name key to sort by (last name, first name)

        :return: str Name sort key
        """

        return f'{self.who["last"]}, {self.who["first"]}'

    @property
    def assigned(self) -> str:
        """
        Builds full name (including title & department) of all assigned

        :return: str Full name(s)
        """

        # If only one assignee ..
        if not self.others:
            # .. provide its name
            return self.name

        # Provide all assignees
        return "; ".join([self.name, format_people(self.others)])


class Dates(ABC):
    """
    Base class for multiple assignments
    """

    # Assignment data
    data: Sequence[Date]

    # Sort order
    sort_order: list[str] = []

    # Event host (ICS only)
    creator: str = "S1SYPHOS"

    # Preferred timezone (ICS only)
    timezone: str = "Europe/Berlin"

    def __iter__(self):  # type: ignore
        """
        Enables iteration
        """

        yield from self.data

    def __len__(self) -> int:
        """
        Enables 'len()'

        :return: int
        """

        return len(self.data)

    def to_dict(self) -> list[dict[str, Any]]:
        """
        Exports assignment data

        :return: list[dict[str, Any]]
        """

        return self.sort(self.data)

    def filter(self, query: list[str] | str) -> list[dict[str, Any]]:
        """
        Filters assignments by search term(s)

        :param query: List[str] | str Search terms
        :return: list[dict[str, Any]] Filtered assignments
        """

        # If query empty ..
        if not query:
            # .. provide original data
            return self.to_dict()

        # If query represents string ..
        if isinstance(query, str):
            # .. make it a list
            query = [query]

        # Create data buffer
        dates: list[dict[str, str]] = []

        # Loop over search terms in order to ..
        for term in query:
            # .. filter out relevant items
            dates.extend(
                [
                    item
                    for item in copy.deepcopy(self.data)
                    if term.lower() in item.name.lower()
                ]
            )

        return self.sort(dates)

    def sort(self, dates: list[Date]) -> list[dict[str, Any]]:
        """
        Sorts dates

        :param dates: list[Date] Unsorted dates
        :return: list[dict[str, Any]] Sorted dates
        """

        # Sort data
        dates.sort(key=attrgetter(*self.sort_order))

        return [item.to_dict() for item in dates]

    @abstractmethod
    def data2ics(self) -> ics.Calendar:
        """
        Exports assignments as ICS calendar object

        :return: ics.Calendar
        """

    # HELPERS

    def to_time(self, date_time: str, fmt: str = "%Y-%m-%d") -> datetime.datetime:
        """
        Datetime object helper

        :param date_time: str Date and/or time
        :param fmt: str Datetime pattern
        :return: datetime.datetime
        """

        return datetime.datetime.strptime(date_time, fmt)

    def add_attendees(self, date: Date, event: ics.Event) -> None:
        """
        Adds people as attendees to event

        :param date: Date Date object
        :param event: ics.Event Attended event
        :return: None
        """

        def add(person: dict[str, str], event: ics.Event):
            """
            Adds person as attendee to event

            :param dict[str, str]: Attending person
            :param event: ics.Event Attended event
            :return: None
            """

            # Build attendee
            attendee = ics.Attendee("")

            # Edit name (= title, full name & department)
            attendee.common_name = format_person(person)

            # Add to assignment
            event.add_attendee(attendee)

        # Add assigned person
        add(date.who, event)

        # Add remaining people
        for person in date.others:
            add(person, event)


def format_person(person: dict[str, str]) -> str:  # type: ignore
    """
    Formats single person

    :param person: dict[str, str] Data representing assigned person
    :return: str Formatted person
    """

    return " ".join(
        [
            string.strip()
            for string in [
                person["title"],
                person["doc"],
                person["first"],
                person["last"],
                person["department"],
            ]
            if string
        ]
    )


def format_people(people: list[dict[str, str]]) -> str:
    """
    Formats assigned people

    :param people: list[dict[str, str]] Data representing assigned people
    :return: str Formatted people
    """

    # Bring people together
    return "; ".join([format_person(person) for person in people])



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

from __future__ import annotations

from dataclasses import dataclass
import datetime
import zoneinfo

import ics  # type: ignore

from . import Date, Dates
from ..utils import data2hash


@dataclass
class ExpressDate(Date):
    """
    Holds express service date
    """

    start: str
    end: str


class ExpressDates(Dates):
    """
    Holds express service dates
    """

    # Express service dates
    data: list[ExpressDate]

    # Sort order
    sort_order: list[str] = ["start", "name2sort", "end"]

    def __init__(self, express_dates: list[dict[str, str]]) -> None:
        """
        Creates 'ExpressDates' instance

        :param express_dates: list[dict[str, str]] Data representing express dates
        :return: None
        """

        self.data = [ExpressDate(**item) for item in express_dates]

    def filter(self, query: list[str] | str) -> ExpressDates:
        """
        Filters court dates by search term(s)

        :param query: List[str] | str Search terms
        :return: ExpressDates Filtered express service dates
        """

        return ExpressDates(super().filter(query))

    def data2ics(self) -> ics.Calendar:
        """
        Exports express service dates as ICS calendar object

        :return: ics.Calendar
        """

        # Create calendar
        calendar = ics.Calendar(creator=self.creator)

        # Define timezone
        timezone = zoneinfo.ZoneInfo(self.timezone)

        # Iterate over assignments
        for item in self.data:
            # Define timezone, date & times
            first_day = self.to_time(item.start)
            last_day = self.to_time(item.end)

            # Create event
            event = ics.Event(
                uid=data2hash(item),
                name="Eildienst (StA)",
                created=datetime.datetime.now(timezone),
                begin=first_day.replace(tzinfo=timezone),
                end=last_day.replace(tzinfo=timezone),
            )

            # Transform to all-day event
            event.make_all_day()

            # Add assignee(s) as attendee(s)
            self.add_attendees(item, event)

            # Add event to calendar
            calendar.events.add(event)

        return calendar



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

from .regex import COURT_DATES, EXPRESS_DATES, PERSON
from .sta import StA

__all__ = [
    # Main class
    "StA",
    # RegExes
    "COURT_DATES",
    "EXPRESS_DATES",
    "PERSON",
]



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

import re


def _get_data(regex: re.Pattern, string: str) -> list[dict[str, str]]:
    """
    Retrieves data

    :param regex: re.Pattern Compiled regex
    :param string: str String
    :return: list[dict[str, str]] Data
    """

    # Retrieve data
    # (1) Match pattern
    # (2) Clean strings
    return [
        {k: v.strip() for k, v in match.groupdict().items()}
        for match in regex.finditer(string)
    ]


COURT_DATES = re.compile(
    r"""
    # (1) .. location (optional)
    (?P<where>(?:.*?)?)\s?
    # (2) .. time of court date
    (?P<time>\d{2}:\s?\d{2})\s
    # (3) .. docket number
    (?P<docket>\d{2,3}\sU?Js\s\d+\/\d{2})\s
    # (4) .. name(s) of prosecutor(s), namely ..
    (?P<assigned>
        (?:
            # (a) .. last name & doctoral degree (optional)
            (?:(?:Dr\.\s)?[\u00C0-\u017F\w-]+)
            # (b) .. department number (optional)
            (?:\s?(?:\([0-9XIV]+\)))?\s?,\s
            # (c) .. first name
            (?:[\u00C0-\u017F\w-]+)\s?,\s
            # (d) .. official title
            (?:
                (?:
                    Ref|JOI|AAAnw|
                    E?(?:O?StA|O?AA)|
                    (?:RR(?:\'in)?aAA)
                )
                (?:\'in)?
                (?:\s\(ba\))?
            )\s?
        )+
    )
    """,
    re.VERBOSE,
)


def get_court_dates(string: str) -> list[dict[str, str]]:
    """
    Retrieves data representing court dates

    :param string: str String representing court dates
    :return: list[dict[str, str]] Data representing court dates
    """

    return _get_data(COURT_DATES, string)


EXPRESS_DATES = re.compile(
    r"""
    # (1) .. start date
    (?P<start>\d{2}\.\d{2}\.\d{4})\s
    # (2) .. hyphen, followed by whitespace
    (?:-\s)
    # (3) .. end date
    (?P<end>\d{2}\.\d{2}\.\d{4})\s
    # (4) .. name(s) of prosecutor(s), namely ..
    (?P<assigned>
        (?:
            # (a) .. last name & doctoral degree (optional)
            (?:(?:Dr\.\s)?[\u00C0-\u017F\w-]+)
            # (b) .. department number (optional)
            (?:\s(?:\([0-9XIV]+\)))?\s?,\s
            # (c) .. first name
            (?:[\u00C0-\u017F\w-]+)\s?,\s
            # (d) .. official title
            (?:
                (?:
                    Ref|JOI|AAAnw|
                    E?(?:O?StA|O?AA)|
                    (?:RR(?:\'in)?aAA)
                )
                (?:\'in)?
                (?:\s\(ba\))?
            )\s?
        )+
    )
    """,
    re.VERBOSE,
)


def get_express_dates(string: str) -> list[dict[str, str]]:
    """
    Retrieves data representing express service dates

    :param string: str String representing express service dates
    :return: list[dict[str, str]] Data representing express service dates
    """

    return _get_data(EXPRESS_DATES, string)


PERSON = re.compile(
    r"""
    # (1) .. doctoral degree (optional)
    (?P<doc>(?:Dr\.)?)\s??
    # (2) .. last name
    (?P<last>[\u00C0-\u017F\w-]+)\s?
    # (3) .. department number (optional)
    (?P<department>(?:\([0-9XIV]+\))?)\s?,\s?
    # (4) .. first name
    (?P<first>[\u00C0-\u017F\w-]+)\s?,\s?
    # (5) .. official title, being either ..
    (?P<title>
        (?:
            # (a) .. Rechtsreferendar:in
            # - Ref / Ref'in
            #
            # (b) .. Justizoberinspektor:in
            # - JOI / JOI'in
            #
            # (c) .. Amtsanwaltsanwärter:in
            # - AAAnw / AAAnw'in
            Ref|JOI|AAAnw|

            # (d) .. (Erste:r / Ober-) Staatsanwalt:anwältin
            # - OStA / OStA'in
            # - EStA / EStA'in
            # - StA / StA'in
            # (e) .. (Erste:r) (Oberamts-) Anwalt:Anwältin
            # - EOAA / EOAA'in
            # - OAA / OAA'in
            E?(?:O?StA|O?AA)|

            # (f) .. Regierungsrat:rätin als Amtsanwalt:anwältin
            # - RRaAA / RR'inaAA'in
            (?:RR(?:\'in)?aAA)
        )
        (?:\'in)?
        (?:\s\(ba\))?
    )
    """,
    re.VERBOSE,
)


def get_people(string: str) -> list[dict[str, str]]:
    """
    Retrieves data representing assigned people

    :param string: str String representing assigned people
    :return: list[dict[str, str]] Data representing people
    """

    return _get_data(PERSON, string)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sitzungsdienst', 'sitzungsdienst.models']

package_data = \
{'': ['*']}

install_requires = \
['ics>=0.7,<0.8', 'pypdf2==2.0.0']

setup_kwargs = {
    'name': 'sitzungsdienst',
    'version': '2.3.4',
    'description': 'Handles weekly assignment PDFs as exported by "web.sta"',
    'long_description': "# sitzungsdienst\n[![License](https://badgen.net/badge/license/GPL/blue)](https://codeberg.org/S1SYPHOS/sitzungsdienst/src/branch/main/LICENSE) [![PyPI](https://badgen.net/pypi/v/sitzungsdienst)](https://pypi.org/project/sitzungsdienst) [![Coverage](https://badgen.net/badge/coverage/100/cyan)](https://codeberg.org/S1SYPHOS/sitzungsdienst/src/branch/main/COVERAGE) [![Build](https://ci.codeberg.org/api/badges/S1SYPHOS/sitzungsdienst/status.svg)](https://codeberg.org/S1SYPHOS/sitzungsdienst/issues)\n\nA simple Python utility for working with weekly assignment PDFs as exported by [`web.sta`](https://www.dvhaus.de/leistungen/web.sta).\n\n\n## Getting started\n\nSimply install all dependencies inside a virtual environment to get started:\n\n```bash\n# Clone repository & change directory\ngit clone https://codeberg.org/S1SYPHOS/sitzungsdienst && cd sitzungsdienst\n\n# Set up & activate virtualenv\npoetry shell\n\n# Install dependencies\npoetry install\n```\n\n\n## Usage\n\nUsing this library is straightforward:\n\n```python\nfrom sitzungsdienst import StA\n\n# Pass file path (or its stream) & retrieve data\ncourt_dates, express_dates = StA.runs('path/to/file.pdf')\n\n# You may also pass multiple file paths (or their streams)\ncourt_dates, express_dates = StA.runs(['path/to/file1.pdf', 'path/to/file2.pdf'])\n\n# Use a subset by filtering it\nfiltered_court = court_dates.filter(['alice', 'bob'])\nfiltered_express = express_dates.filter('john')\n\n# Get iCalendar data\nics = filtered_court.data2ics()\nprint(ics)\n\n##\n# BEGIN:VCALENDAR\n# VERSION:2.0\n# ..\n# ..\n```\n\n**Note**: Since all data methods are using `async`, you should either `await` them (inside your own `asyncio` context) or call them using `asyncio.run()` (see below).\n\n```python\nimport asyncio\nfrom sitzungsdienst import StA\n\nasync def main(file):\n    # ..\n\n    return await StA.run(file)\n\n# Retrieve data\ndata = asyncio.gather(*[main(file) for file in files])\n\n# ..\n```\n\nIf you want to see it in action, head over to the [`sitzungsapp`](https://codeberg.org/S1SYPHOS/sitzungsapp)!\n\n\n**Happy coding!**\n",
    'author': 'Digitalbüro',
    'author_email': 'pypi@digitalbuero.eu',
    'maintainer': 'Martin Folkers',
    'maintainer_email': 'hello@twobrain.io',
    'url': 'https://digitalbuero.eu',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
This module is part of the 'sitzungsdienst' package,
which is released under GPL-3.0-only license.
"""

from hashlib import blake2b
from typing import Any, Iterable


def data2hash(data: Any) -> str:
    """
    Builds hash over given data

    :param data: typing.Any Data
    :return: str Hash
    """

    return blake2b(str(data).encode("utf-8"), usedforsecurity=False).hexdigest()


def flatten(data: Iterable[Iterable[Any]]) -> list[Any]:
    """
    Flattens list of lists

    :param data: typing.Iterable[typing.Iterable[typing.Any]] List of lists
    :return: list Flattened list
    """

    return [item for sublist in data for item in sublist]


def sort_din5007(string: str) -> tuple[str, str]:
    """
    Sorts strings with german umlauts according to DIN 5007 ('sorted' helper)

    :param string: str Input string
    :return: tuple Strings for comparison
    """

    # Lowercase string
    lowercase = string.lower()

    # Replace umlauts
    lowercase = lowercase.replace("ä", "a")
    lowercase = lowercase.replace("ö", "o")
    lowercase = lowercase.replace("ü", "u")
    lowercase = lowercase.replace("ß", "ss")

    return (lowercase, string.swapcase())
